PROGRAM DNAML(INFILE, OUTPUT);
   (* VERSION 2.1 -- (old version, slightly updated)  *)
CONST MAXSP  =  15; (* MAXIMUM NUMBER OF SPECIES *)
   MAXSP2    =  27; (* 2 X MAXSP - 3 *)
   MAXSP1    =  28; (* MAXSP2 + 1 *)
   MAXSITE   =  60; (* MAXIMUM NUMBER OF BASE SITES PER SEQUENCE *)
   BONESIZE  =  481; (*ONE PLUS EIGHT TIMES MAXSITE*)
   TRAVERSE  =  7; (* NUMBER OF PASSES THROUGH TREE IN SMOOTHING *)
   ITERATE   =  5; (* NUMBER OF ITERATIONS OF P PER PASS IN EACH SEGMENT *)
   NAMESIZE  =  10; (* LENGTH OF SPECIES NAME *)
TYPE  BASE    =  (A, C, G, T);
   NAYM    =  ARRAY[1..NAMESIZE] OF CHAR;
   LHOODARRAY = ARRAY[A..T] OF ARRAY[1..MAXSITE] OF REAL;
   NODENUM = 0..MAXSP1;
   SITENUM = 1..MAXSITE;
   VIEW    =  RECORD
              NAME            :  NAYM;
              X               :  LHOODARRAY;
              END;
   EDGEPTR   =  ^EDGE;
   NODEPTR   =  ^NODE;
   EDGE    =  RECORD
               FRONT, BACK     :  NODEPTR;
               ARRANGED        :  BOOLEAN;
               BACKVIEW,
               FRONTVIEW       :  VIEW;
               P               :  REAL;
               END;
   NODE      =  RECORD
                INDEX           :  NODENUM;
                ANCESTOR,
                FORKLEFT,
                FORKRIGHT       :  EDGEPTR;
                END;
   INSERTION =  RECORD
                WHERE           :  EDGEPTR;
                LIKLI           :  REAL;
                END;
   BONE      =  RECORD
                X, Y            :  LHOODARRAY;
                P               :  REAL;
                END;
VAR INFILE           : TEXT;
    TREE             : EDGEPTR;
    NUMSP, I         : INTEGER;
    NUMSITE          : SITENUM;
    PI               : ARRAY[A..T] OF REAL;
PROCEDURE MAKETREE (VAR PHENOFILE :  TEXT);
   VAR Q, R, STARTER  :  EDGEPTR;
       PHENO    :  VIEW;
       TEMPLIKLI   :  REAL;
       INCOMPLETE   :  BOOLEAN;
       BESTREE     :  INSERTION;
       BONEINDEX, INDEXER : NODENUM;
       TOTALTREES : INTEGER;
       CLOSET      :  ARRAY[1..MAXSP2] OF BONE;
       TABLE1, TABLE2 :  ARRAY[1..MAXSITE] OF REAL;
       U        :  BASE;
   PROCEDURE UPDATE (STEM : EDGEPTR);
      VAR SUM1, SUM2, SIGMA : REAL;
          OTHER : EDGEPTR;
      PROCEDURE STEP1;
            VAR K    : SITENUM;
                U, B   : BASE;
         BEGIN(*STEP1*)
              (*RECONSTRUCT FRONTVIEW OF SEGMENT*)
            IF STEM^.FRONT^.FORKLEFT<>NIL
            THEN WITH STEM^ DO BEGIN
               FOR K:=1 TO NUMSITE DO BEGIN
                  FOR U:=A TO T DO BEGIN
                     SUM1:=0;
                     SUM2:=0;
                     FOR B:=A TO T DO BEGIN
                        SUM1:=SUM1+PI[B]*
                              FRONT^.FORKLEFT^.FRONTVIEW.X[B][K];
                        SUM2:=SUM2+PI[B]*
                           FRONT^.FORKRIGHT^.FRONTVIEW.X[B][K];
                        END;
                     SUM1:=FRONT^.FORKLEFT^.P*SUM1+
                        (1-FRONT^.FORKLEFT^.P)*
                        FRONT^.FORKLEFT^.FRONTVIEW.X[U][K];
                     SUM2:=FRONT^.FORKRIGHT^.P*SUM2+
                        (1-FRONT^.FORKRIGHT^.P)*
                         FRONT^.FORKRIGHT^.FRONTVIEW.X[U][K];
                         FRONTVIEW.X[U][K]:=SUM1*SUM2;
                     END;
                  END;
               END;
         END; (*STEP1*)
      PROCEDURE STEP2;
         VAR  K    : SITENUM;
              U, B   : BASE;
         BEGIN(*STEP 2*)
              (*RECONSTRUCT BACKVIEW OF SEGMENT*)
            IF STEM^.BACK^.ANCESTOR<>NIL
            THEN BEGIN
               IF STEM^.BACK^.FORKRIGHT=STEM
               THEN OTHER:=STEM^.BACK^.FORKLEFT
               ELSE OTHER:=STEM^.BACK^.FORKRIGHT;
               WITH STEM^ DO BEGIN
                  FOR K:=1 TO NUMSITE DO BEGIN
                     FOR U:=A TO T DO BEGIN
                        SUM1:=0;
                        SUM2:=0;
                        FOR B:=A TO T DO BEGIN
                           SUM1:=SUM1+PI[B]*
                              BACK^.ANCESTOR^.BACKVIEW.X[B][K];
                           SUM2:=SUM2+PI[B]*OTHER^.FRONTVIEW.X[B][K];
                           END;
                        SUM1:=BACK^.ANCESTOR^.P*SUM1+
                           (1-BACK^.ANCESTOR^.P)
                           *BACK^.ANCESTOR^.BACKVIEW.X[U][K];
                        SUM2:=OTHER^.P*SUM2+(1-OTHER^.P)*
                           OTHER^.FRONTVIEW.X[U][K];
                        BACKVIEW.X[U][K]:=SUM1*SUM2;
                        END;
                     END;
                  END;
               END;
         END; (*STEP2*)
      PROCEDURE STEP3;
         VAR  K  : SITENUM;
              B  : BASE;
         BEGIN(*STEP 3*)
               (*CONSTRUCT TABLE1 AND TABLE2 ARRAYS*)
            FOR K:=1 TO NUMSITE DO BEGIN
               TABLE1[K]:=0;
               SUM1:=0;
               SUM2:=0;
               FOR B:=A TO T DO BEGIN
                  TABLE1[K]:=TABLE1[K]+PI[B]*
                      STEM^.FRONTVIEW.X[B][K]*
                      STEM^.BACKVIEW.X[B][K];
                  SUM1:=SUM1+PI[B]*
                     STEM^.FRONTVIEW.X[B][K];
                  SUM2:=SUM2+PI[B]*
                     STEM^.BACKVIEW.X[B][K];
                  END;
               TABLE2[K]:=SUM1*SUM2;
               END;
         END; (*STEP 3*)
      PROCEDURE STEP4;
         VAR  COUNT : INTEGER;
              K : SITENUM;
         BEGIN(*STEP4*)
            FOR COUNT:=1 TO ITERATE DO BEGIN
               SIGMA:=0;
               IF STEM^.P>0.00001
               THEN BEGIN
                  FOR K:=1 TO NUMSITE DO
                     SIGMA:=SIGMA+TABLE2[K]*STEM^.P/
                      (TABLE1[K]*(1-STEM^.P)+TABLE2[K]*STEM^.P);
                     STEM^.P:=SIGMA/NUMSITE;
                  END;
               END;
         END; (*STEP4*)
      BEGIN(*UPDATE*)
         STEP1;
         STEP2;
         STEP3;
         STEP4;
      END; (*UPDATE*)
   PROCEDURE SMOOTH(SGMNT : EDGEPTR);
      BEGIN (*SMOOTH *)
         UPDATE(SGMNT);
         IF SGMNT^.FRONT^.FORKLEFT<>NIL
         THEN SMOOTH(SGMNT^.FRONT^.FORKLEFT);
         IF SGMNT^.FRONT^.FORKRIGHT<>NIL
         THEN SMOOTH(SGMNT^.FRONT^.FORKRIGHT);
      END;(*SMOOTH*)
   PROCEDURE STORE(STEM : EDGE);
      BEGIN(*STORE*)
         BONEINDEX:=BONEINDEX+1;
         WITH STEM DO BEGIN
            CLOSET[BONEINDEX].X:=FRONTVIEW.X;
            CLOSET[BONEINDEX].Y:=BACKVIEW.X;
            CLOSET[BONEINDEX].P:=P;
            END;
         IF STEM.FRONT^.FORKLEFT<>NIL
         THEN STORE(STEM.FRONT^.FORKLEFT^);
         IF STEM.FRONT^.FORKRIGHT<>NIL
         THEN STORE(STEM.FRONT^.FORKRIGHT^);
      END;(*STORE*)
    PROCEDURE GETREE(VAR STEM : EDGE);
       BEGIN(*GETREE*)
          BONEINDEX:=BONEINDEX+1;
          WITH STEM DO BEGIN
            FRONTVIEW.X:=CLOSET[BONEINDEX].X;
            BACKVIEW.X:=CLOSET[BONEINDEX].Y;
            P:=CLOSET[BONEINDEX].P;
            END;
         IF STEM.FRONT^.FORKLEFT<>NIL
         THEN GETREE(STEM.FRONT^.FORKLEFT^);
         IF STEM.FRONT^.FORKRIGHT<>NIL
         THEN GETREE(STEM.FRONT^.FORKRIGHT^);
      END;(*GETREE*)
   FUNCTION EVALUATE(STEM : EDGEPTR)   :   REAL;
      VAR K   :   INTEGER;
            B         :   BASE;
            TEMPLIKLI, SUM1, SUM2, SUM3    :   REAL;
      FUNCTION LGNATURAL ( X : REAL ) : REAL;
         VAR Y : REAL;
         BEGIN (*LGNATURAL*)
            Y := LN(X);
            (* IF X<0 THEN Y := -ABS(Y);  -- MAY NEED THIS IF OLD UCSD PASCAL*)
            LGNATURAL := Y;
         END; (*LGNATURAL*)
      BEGIN (*EVALUATE*)
         TEMPLIKLI:=0;
         WITH STEM^ DO BEGIN
            FOR K:=1 TO NUMSITE DO BEGIN
               SUM1:=0;
               SUM2:=0;
               SUM3:=0;
               FOR B:=A TO T DO BEGIN
                  SUM1:=SUM1+PI[B]*
                      FRONTVIEW.X[B][K]*
                      BACKVIEW.X[B][K];
                  SUM2:=SUM2+PI[B]*
                      FRONTVIEW.X[B][K];
                  SUM3:=SUM3+PI[B]*
                      BACKVIEW.X[B][K];
                  END;
               TEMPLIKLI:=TEMPLIKLI+LGNATURAL((1.0-P)*
                      SUM1+P*SUM2*SUM3);
               END;
            END;
         EVALUATE:=TEMPLIKLI;
      END; (*EVALUATE*)
   PROCEDURE CLEAR(VAR V : VIEW);
      VAR K      : SITENUM;
          B      : BASE;
      BEGIN(*CLEAR*)
         FOR K:=1 TO NUMSITE DO BEGIN
            FOR B:=A TO T DO
                  V.X[B][K]:=0;
            END;
      END;(*CLEAR*)
   PROCEDURE GRAB(VAR PHENO : VIEW);
      VAR COUNT : INTEGER;
          B   : CHAR;
          I : INTEGER;
      BEGIN(*GRAB*)
         WRITE('   ');
         COUNT := 1;
         WHILE COUNT <=NUMSITE DO BEGIN
            IF EOLN(INFILE) THEN READLN(INFILE);
            READ(INFILE,B);
            IF ORD(B)>95 THEN B := CHR(ORD(B)-32); (*ASSUMES ASCII CODE*)
            CASE B OF
               'N', 'X', '?', 'O', ' ' : BEGIN
                      PHENO.X[A][COUNT]:=1;
                      PHENO.X[C][COUNT]:=1;
                      PHENO.X[G][COUNT]:=1;
                      PHENO.X[T][COUNT]:=1;
                     END;
               'A' : PHENO.X[A][COUNT]:=1;
               'C' : PHENO.X[C][COUNT]:=1;
               'G' : PHENO.X[G][COUNT]:=1;
               'U','T' : PHENO.X[T][COUNT]:=1;
               'H' : BEGIN
                     PHENO.X[A][COUNT] := 1;
                     PHENO.X[C][COUNT] := 1;
                     PHENO.X[U][COUNT] := 1;
                     END;
               'R' : BEGIN
                     PHENO.X[A][COUNT] := 1;
                     PHENO.X[G][COUNT] := 1;
                     END;
               'Y' : BEGIN
                     PHENO.X[C][COUNT] := 1;
                     PHENO.X[U][COUNT] := 1;
                     END;
               END;
            COUNT := COUNT + 1;
            WRITE(B);
            IF (COUNT MOD 48) = 1
            THEN BEGIN
                WRITELN;
                FOR I := 1 TO NAMESIZE DO WRITE(' ');
                WRITE('   ');
                END;
            IF (COUNT MOD 3) = 1 THEN WRITE(' ');
            END;
         READLN(INFILE);
         WRITELN;
      END; (*GRAB*)
   PROCEDURE BUILDQR;
      VAR SGMNT : EDGEPTR;
          TRIANGLE : NODEPTR;
          COUNT : INTEGER;
      BEGIN (*BUILDQR*)
         NEW(SGMNT);
         NEW(TRIANGLE);
         INDEXER:=INDEXER+1;
         TRIANGLE^.INDEX:=INDEXER;
         WRITE(' ');
         FOR COUNT:=1 TO NAMESIZE DO BEGIN
            READ(INFILE,PHENO.NAME[COUNT]);
            WRITE(PHENO.NAME[COUNT]);
            SGMNT^.BACKVIEW.NAME[COUNT]:=' ';
            END;
         CLEAR(PHENO);
         GRAB(PHENO);
         SGMNT^.FRONTVIEW:=PHENO;
         SGMNT^.FRONT:=TRIANGLE;
         TRIANGLE^.ANCESTOR:=SGMNT;
         TRIANGLE^.FORKLEFT:=NIL;
         TRIANGLE^.FORKRIGHT:=NIL;
         NEW(TRIANGLE);
         INDEXER:=INDEXER+1;
         TRIANGLE^.INDEX:=INDEXER;
         TRIANGLE^.FORKLEFT:=SGMNT;
         SGMNT^.BACK:=TRIANGLE;
         NEW(SGMNT);
         FOR COUNT:=1 TO NAMESIZE DO BEGIN
            SGMNT^.FRONTVIEW.NAME[COUNT]:=' ';
            SGMNT^.BACKVIEW.NAME[COUNT]:=' ';
            END;
         TRIANGLE^.FORKRIGHT:=SGMNT;
         TRIANGLE^.ANCESTOR:=NIL;
         SGMNT^.FRONT:=NIL;
         SGMNT^.BACK:=TRIANGLE;
         Q:=TRIANGLE^.FORKLEFT;
         R:=TRIANGLE^.FORKRIGHT;
      END; (*BUILDQR*)
   PROCEDURE BUILDSIMPLETREE;
      VAR SGMNT      :    EDGEPTR;
          TRIANGLE :    NODEPTR;
          COUNT      :    INTEGER;
      BEGIN (*BUILDSIMPLETREE*)
         INDEXER:=0;
         BUILDQR;
         NEW (SGMNT);
         NEW (TRIANGLE);
         INDEXER:=INDEXER+1;
         TRIANGLE^.INDEX:=INDEXER;
         TRIANGLE^.ANCESTOR:=NIL;
         TRIANGLE^.FORKRIGHT:=NIL;
         SGMNT^.BACK:=TRIANGLE;
         SGMNT^.FRONT:=Q^.BACK;
         FOR COUNT:=1 TO NAMESIZE DO
               SGMNT^.FRONTVIEW.NAME[COUNT]:=' ';
         Q^.BACK^.ANCESTOR:=SGMNT;
         WRITE(' ');
         FOR COUNT:=1 TO NAMESIZE DO BEGIN
            READ(INFILE,PHENO.NAME[COUNT]);
            WRITE(PHENO.NAME[COUNT]);
            END;
         CLEAR(PHENO);
         GRAB(PHENO);
         SGMNT^.BACKVIEW:=PHENO;
         TRIANGLE^.FORKLEFT:=SGMNT;
         NEW(TRIANGLE);
         INDEXER:=INDEXER+1;
         TRIANGLE^.INDEX:=INDEXER;
         R^.FRONT:=TRIANGLE;
         TRIANGLE^.ANCESTOR:=R;
         WRITE(' ');
         FOR COUNT:=1 TO NAMESIZE DO BEGIN
            READ(INFILE,PHENO.NAME[COUNT]);
            WRITE(PHENO.NAME[COUNT]);
            END;
         CLEAR(PHENO);
         GRAB(PHENO);
         R^.FRONTVIEW:=PHENO;
         TRIANGLE^.FORKLEFT:=NIL;
         TRIANGLE^.FORKRIGHT:=NIL;
         SGMNT^.P:=0.5;
         R^.P:=0.5;
         R^.BACK^.FORKLEFT^.P:=0.5;
         TREE:=SGMNT;
                  (*SET TREE TO POINT TO "QUASI-ANCESTRAL
                   SGMNT:   RECURSION ROUTINES
                   NEED SOMEWHERE TO BEGIN AND
                   THIS SOMEWHERE IS UNIQUE BECAUSE
                   OF THE POLARITY OF THE SEGMENTS*)
         FOR COUNT:=1 TO 2 DO SMOOTH(TREE);
      END; (*BUILDSIMPLETREE*)
   PROCEDURE CHECK(STEM : EDGEPTR);
      BEGIN (*CHECK*)
         IF STEM^.P>0.9999 THEN STEM^.P:=0.9;
         IF STEM^.P<0.0001 THEN STEM^.P:=0.1;
         IF STEM^.FRONT^.FORKLEFT<>NIL
         THEN CHECK(STEM^.FRONT^.FORKLEFT);
         IF STEM^.FRONT^.FORKRIGHT<>NIL
         THEN CHECK(STEM^.FRONT^.FORKRIGHT);
      END; (*CHECK*)
   PROCEDURE INSERT (P, R   : EDGEPTR);
      BEGIN (*INSERT*)
         R^.FRONT:=P^.FRONT;
         P^.FRONT:=R^.BACK;
         R^.BACK^.ANCESTOR:=P;
         R^.FRONT^.ANCESTOR:=R;
         R^.FRONTVIEW:=P^.FRONTVIEW;
         R^.ARRANGED:=P^.ARRANGED;
         R^.P:=P^.P/2;
         P^.P:=R^.P;
         R^.BACK^.FORKLEFT^.P:=0.5;
         UPDATE(R);
         UPDATE(P);
         UPDATE(R^.BACK^.FORKLEFT);
      END; (*INSERT*)
   PROCEDURE DELETE(R : EDGEPTR);
      VAR P : EDGEPTR;
      BEGIN (*DELETE*)
         P:=R^.BACK^.ANCESTOR;
         P^.FRONT:= R^.FRONT;
         R^.FRONT:=NIL;
         P^.P:=0.5;
         P^.FRONT^.ANCESTOR:=P;
         R^.BACK^.ANCESTOR:=NIL;
         P^.FRONTVIEW:=R^.FRONTVIEW;
         UPDATE(P);
         IF P^.FRONT^.FORKLEFT<>NIL
         THEN BEGIN
            UPDATE(P^.FRONT^.FORKLEFT);
            UPDATE(P^.FRONT^.FORKRIGHT);
            END;
      END; (*DELETE*)
   PROCEDURE PLACENEWTIP;
      VAR COUNT : INTEGER;
      PROCEDURE SEARCH (STEM : EDGEPTR);
         VAR COUNT   :   INTEGER;
         BEGIN (*SEARCH*)
            IF STEM^.FRONT^.FORKLEFT<>NIL
            THEN BEGIN
               TOTALTREES:=TOTALTREES+1;
               BONEINDEX:=0;
               GETREE(TREE^);
               CHECK(TREE);
               INSERT (STEM^.FRONT^.FORKLEFT, R);
               FOR COUNT:=1 TO TRAVERSE DO SMOOTH(TREE);
               TEMPLIKLI:=EVALUATE(TREE);
               DELETE(R);
               IF TEMPLIKLI>BESTREE.LIKLI
               THEN BEGIN
                  BESTREE.WHERE:=STEM^.FRONT^.FORKLEFT;
                  BESTREE.LIKLI:=TEMPLIKLI;
                  END;
               SEARCH (STEM^.FRONT^.FORKLEFT);
               END;
            IF STEM^.FRONT^.FORKRIGHT<>NIL
            THEN BEGIN
               TOTALTREES:=TOTALTREES+1;
               BONEINDEX:=0;
               GETREE(TREE^);
               CHECK(TREE);
               INSERT (STEM^.FRONT^.FORKRIGHT, R);
               FOR COUNT:=1 TO TRAVERSE DO SMOOTH(TREE);
               TEMPLIKLI:=EVALUATE(TREE);
               DELETE(R);
               IF TEMPLIKLI>BESTREE.LIKLI
               THEN BEGIN
                  BESTREE.WHERE:=STEM^.FRONT^.FORKRIGHT;
                  BESTREE.LIKLI:=TEMPLIKLI;
                  END;
               SEARCH (STEM^.FRONT^.FORKRIGHT);
               END;
         END; (*SEARCH*)
      BEGIN (*PLACENEWTIP*)
         BONEINDEX:=0;
         GETREE(TREE^);
         CHECK(TREE);
         INSERT(TREE, R);
         TOTALTREES:=TOTALTREES+1;
         FOR COUNT:=1 TO TRAVERSE DO SMOOTH(TREE);
         TEMPLIKLI:=EVALUATE(TREE);
         DELETE(R);
         BESTREE.WHERE:=TREE;
         BESTREE.LIKLI:=TEMPLIKLI;
         SEARCH(TREE);
         CHECK(TREE);
         INSERT(BESTREE.WHERE, R);
         FOR COUNT:=1 TO TRAVERSE DO SMOOTH(TREE);
         BONEINDEX:=0;
         STORE(TREE^);
      END;(*PLACENEWTIP*)
   PROCEDURE INITIALIZE (STEM : EDGEPTR);
      BEGIN (*INITIALIZE*)
         STEM^.ARRANGED:=FALSE;
         IF STEM^.FRONT^.FORKLEFT<>NIL
         THEN INITIALIZE (STEM^.FRONT^.FORKLEFT);
         IF STEM^.FRONT^.FORKRIGHT<>NIL
         THEN INITIALIZE (STEM^.FRONT^.FORKRIGHT);
      END; (*INITIALIZE*)
   PROCEDURE REARRANGE(STEM   :   EDGEPTR);
      VAR RERANGEABLE    : BOOLEAN;
            REAR      : EDGEPTR;
            POSITION,
            RO, COUNT   : INTEGER;
      PROCEDURE ROTATE(VAR STEM : EDGEPTR);
         VAR   RIGHT      : BOOLEAN;
            TEMP,
            ROOT   : EDGEPTR;
         BEGIN
            ROOT:=STEM^.BACK^.ANCESTOR;
            RIGHT:=FALSE;
            IF STEM^.BACK^.FORKRIGHT=STEM
            THEN RIGHT:=TRUE;
            IF NOT RIGHT
            THEN BEGIN
               TEMP:=ROOT^.FRONT^.FORKLEFT;
               ROOT^.FRONT^.FORKLEFT:=ROOT^.FRONT^.FORKRIGHT;
               ROOT^.FRONT^.FORKRIGHT:=TEMP;
               END;
            DELETE(STEM);
            INSERT(ROOT^.FRONT^.FORKLEFT,STEM);
            STEM:=ROOT^.FRONT^.FORKLEFT;
            IF RIGHT
            THEN BEGIN
               TEMP:=ROOT^.FRONT^.FORKLEFT;
               ROOT^.FRONT^.FORKLEFT:=ROOT^.FRONT^.FORKRIGHT;
               ROOT^.FRONT^.FORKRIGHT:=TEMP;
               END;
          END; (*ROTATE*)
      BEGIN (*REARRANGE*)
         RERANGEABLE:=FALSE;
         IF (STEM^.FRONT^.FORKLEFT<>NIL) THEN RERANGEABLE:=TRUE;
         IF RERANGEABLE
         THEN BEGIN
            POSITION:=1;
            FOR RO:=1 TO 2 DO BEGIN
               TOTALTREES:=TOTALTREES+1;
               ROTATE(STEM);
               UPDATE(STEM);
               CHECK(TREE);
               FOR COUNT:=1 TO TRAVERSE DO SMOOTH(TREE);
               TEMPLIKLI:=EVALUATE(TREE);
               IF TEMPLIKLI>BESTREE.LIKLI
               THEN BEGIN
                  BONEINDEX:=0;
                  STORE(TREE^);
                  POSITION:=RO+1;
                  BESTREE.LIKLI:=TEMPLIKLI;
                  END;
               END;
            FOR RO:=1 TO POSITION DO ROTATE(STEM);
            IF POSITION<>1 THEN INCOMPLETE:=TRUE;
            BONEINDEX:=0;
            GETREE(TREE^);
            STEM^.ARRANGED:=TRUE;
            IF STEM^.BACK^.FORKRIGHT=STEM
            THEN REAR:=STEM^.BACK^.FORKLEFT
            ELSE REAR:=STEM^.BACK^.FORKRIGHT;
            IF (NOT STEM^.FRONT^.FORKLEFT^.ARRANGED)
            THEN REARRANGE (STEM^.FRONT^.FORKLEFT);
            IF (NOT STEM^.FRONT^.FORKRIGHT^.ARRANGED)
            THEN REARRANGE (STEM^.FRONT^.FORKRIGHT);
            IF (NOT REAR^.ARRANGED)
            THEN REARRANGE (REAR);
            END
         ELSE STEM^.ARRANGED:=TRUE;
      END;   (*REARRANGE*)
   PROCEDURE PRINTREE( T : NODEPTR );
      (*FELSENSTEIN ALGORITHM FOR PRINTING
        CLADOGRAMS ON A LINE PRINTER*)
      CONST C1=2;
         C2=3;
      VAR HORIZ  :  ARRAY[1..100] OF INTEGER;
          ALPHA  :  ARRAY[1..100] OF CHAR;
          SPECIAL, SPECIAL1  : NAYM;
          I, PAST, INDX, OLDIND, PLACE1, PLACE2 : INTEGER;
      PROCEDURE MIXORDER (R : NODEPTR);
         VAR I, J : INTEGER;
         BEGIN (*MIXORDER*)
            SPECIAL1:=R^.ANCESTOR^.FRONTVIEW.NAME;
            IF R^.FORKLEFT<>NIL
            THEN MIXORDER(R^.FORKLEFT^.FRONT);
            INDX:=R^.INDEX;
            IF R^.FORKLEFT=NIL
            THEN BEGIN
               IF PAST>0
               THEN BEGIN
                  WRITE(' ');
                  FOR I:=1 TO C2*(PAST-1)+1 DO
                        WRITE(ALPHA[I]);
                  FOR I:=1 TO NAMESIZE DO
                        WRITE(SPECIAL[I]);
                  WRITELN;
                  FOR I:=C2*PAST DOWNTO 2 DO BEGIN
                     IF ALPHA[I]='-'
                     THEN ALPHA[I]:=' ';
                     IF ALPHA[I] IN ['0'..'9']
                     THEN BEGIN
                        IF ALPHA[I-1]=' '
                        THEN ALPHA[I]:='!'
                        ELSE ALPHA[I]:=' ';
                        END;
                     END;
                  IF ALPHA[1] IN ['0'..'9'] THEN ALPHA[1]:='!';
                  FOR I:=1 TO C1-1 DO BEGIN
                     WRITE(' ');
                     FOR J:=1 TO C2*PAST DO
                           WRITE(ALPHA[J]);
                     WRITELN;
                     END;
                  END;
               OLDIND:=INDX;
               ALPHA[C2*PAST+1]:='!';
               FOR I:=C2*PAST+2 TO C2*(PAST+1) DO
                     ALPHA[I]:=' ';
               PAST:=PAST+1;
               PLACE2:=PAST;
               HORIZ[INDX]:=PAST;
               SPECIAL:=SPECIAL1;
               END;
            IF R^.FORKLEFT<>NIL
            THEN BEGIN
               HORIZ[INDX]:=HORIZ[R^.FORKLEFT^.FRONT^.INDEX];
               MIXORDER(R^.FORKRIGHT^.FRONT);
               END;
            IF R^.ANCESTOR<>NIL
            THEN BEGIN
               IF R^.ANCESTOR=R^.ANCESTOR^.BACK^.FORKRIGHT
               THEN BEGIN
                  PLACE1:=HORIZ[R^.ANCESTOR^.BACK^.INDEX];
                  FOR I:=C2*(PLACE1-1)+2 TO C2*(PLACE2-1)+1 DO
                  IF NOT (ALPHA[I] IN ['0'..'9']) THEN ALPHA[I]:='-';
                  IF TRUNC(R^.ANCESTOR^.BACK^.INDEX/10)<>0
                  THEN BEGIN
                     ALPHA[C2*(PLACE1-1)+1]:=CHR(ORD('0')+
                       R^.ANCESTOR^.BACK^.INDEX DIV 10);
                     ALPHA[C2*(PLACE1-1)+2]:=CHR(ORD('0')+
                       R^.ANCESTOR^.BACK^.INDEX MOD 10);
                     END
                  ELSE ALPHA[C2*(PLACE1-1)+1]:=CHR(ORD('0')+
                     R^.ANCESTOR^.BACK^.INDEX);
               PLACE2:= PLACE1;
               END;
            END;
         END; (*MIXORDER*)
      BEGIN (*PRINTREE*)
         FOR I:=1 TO NAMESIZE DO
            SPECIAL[I]:=' ';
         PAST:=0;
         WRITELN;
         MIXORDER(T);
         WHILE T^.FORKLEFT<>NIL DO
            T:=T^.FORKRIGHT^.FRONT;
         WRITE(' ');
         SPECIAL:=T^.ANCESTOR^.FRONTVIEW.NAME;
         FOR I:=1 TO C2*(PAST-1)+1 DO WRITE(ALPHA[I]);
         FOR I:=1 TO NAMESIZE DO
            WRITE(SPECIAL[I]);
         WRITELN;
         WRITELN(' !');
         WRITE(' !');
         FOR I:=1 TO NAMESIZE DO
            WRITE(TREE^.BACKVIEW.NAME[I]);
         WRITELN;
      END; (*PRINTREE*)
   PROCEDURE DISPLAY( STEM : EDGEPTR);
      VAR COUNT : INTEGER;
      PROCEDURE CONFLIM (STEM : EDGEPTR);
         VAR B : BASE;
             COUNT : SITENUM;
             SUM1, SUM2, SUM3, V : REAL;
         BEGIN (*CONFLIM*)
            WITH STEM^ DO BEGIN
               V := 0.0;
               FOR COUNT := 1 TO NUMSITE DO BEGIN
                  SUM1 := 0.0;
                  SUM2 := 0.0;
                  SUM3 := 0.0;
                  FOR B := A TO T DO BEGIN
                     SUM1 := SUM1 + PI[B] * FRONTVIEW.X[B, COUNT];
                     SUM2 := SUM2 + PI[B] * BACKVIEW.X[B, COUNT];
                     SUM3 := SUM3 + PI[B] * FRONTVIEW.X[B, COUNT]
                                          * BACKVIEW.X[B, COUNT];
                     END;
                  V := V + SQR((SUM1*SUM2-SUM3)/(SUM1*SUM2*P + SUM3*(1.0-P)));
                  END;
               IF ABS(V) > 1.0E-6
               THEN V := SQRT(1.0/V)
               ELSE V := 1.0E+6;
               WRITELN('   (',P-1.96*V:8:4,',  ',P+1.96*V:8:4,')');
               END;
         END; (*CONFLIM*)
      BEGIN (*DISPLAY*)
         IF STEM^.FRONT^.FORKLEFT=NIL
         THEN BEGIN
            WRITE(' ');
            WRITE(STEM^.BACK^.INDEX:2);
            WRITE('             ');
            FOR COUNT:=1 TO NAMESIZE DO
               WRITE(STEM^.FRONTVIEW.NAME[COUNT]);
            FOR COUNT:=1 TO 15-NAMESIZE DO
               WRITE(' ');
            WRITE(STEM^.P:7:4);
            END
         ELSE BEGIN
            IF STEM^.BACK^.ANCESTOR=NIL
            THEN BEGIN
               WRITE(' ');
               FOR COUNT:=1 TO NAMESIZE DO
                  WRITE(STEM^.BACKVIEW.NAME[COUNT]);
               FOR COUNT:=1 TO 15-NAMESIZE DO
                  WRITE(' ');
               WRITE(STEM^.FRONT^.INDEX:2);
               WRITE('             ');
               WRITE(STEM^.P:7:4);
               END;
               END;
            IF (STEM^.FRONT^.FORKLEFT<>NIL) AND (STEM^.BACK^.ANCESTOR<>NIL)
            THEN BEGIN
               WRITE(' ');
               WRITE(STEM^.BACK^.INDEX:2);
               WRITE('             ');
               WRITE(STEM^.FRONT^.INDEX:2);
               WRITE('             ');
               WRITE(STEM^.P:7:4);
               END;
            CONFLIM(STEM);
            IF STEM^.FRONT^.FORKLEFT<>NIL THEN
               DISPLAY(STEM^.FRONT^.FORKLEFT);
            IF STEM^.FRONT^.FORKRIGHT<>NIL THEN
               DISPLAY(STEM^.FRONT^.FORKRIGHT);
         END; (*DISPLAY*)
   BEGIN (*MAKETREE*)
      READLN(INFILE, NUMSP, NUMSITE);
      FOR U:=A TO T DO
         READ(INFILE,PI[U]);
      READLN(INFILE);
      BUILDSIMPLETREE;
      TOTALTREES:=1;
      BONEINDEX:=0;
      STORE(TREE^);
      BUILDQR;
      PLACENEWTIP;
      PRINTREE(TREE^.FRONT);
      BONEINDEX:=0;
      GETREE(TREE^);
      WRITELN;
      WRITE(' FROM           TO');
      WRITELN('                P         CONFIDENCE LIMITS');
      WRITE(' ----           --');
      WRITELN('               ---        ---------- ------');
      DISPLAY(TREE);
      WRITELN;
      WRITELN('    LOG LIKELIHOOD = ',BESTREE.LIKLI:9:4);
      I := 4;
      WHILE (I<NUMSP) DO BEGIN
          WRITELN;
          WRITELN;
          WRITELN(' READ IN NEXT SPECIES:');
         BUILDQR;
         I := I + 1;
         PLACENEWTIP;
         INCOMPLETE:=TRUE;
         WHILE INCOMPLETE DO BEGIN
            INCOMPLETE:=FALSE;
            INITIALIZE (TREE);
            STARTER:=TREE^.FRONT^.FORKLEFT;
            IF STARTER^.FRONT^.FORKLEFT=NIL
            THEN STARTER:=TREE^.FRONT^.FORKRIGHT;
            REARRANGE (STARTER);
            END;
         PRINTREE(TREE^.FRONT);
         BONEINDEX:=0;
         GETREE(TREE^);
         WRITELN;
         WRITE(' FROM           TO');
         WRITELN('                P         CONFIDENCE LIMITS');
         WRITE(' ----           --');
         WRITELN('               ---        ---------- ------');
         DISPLAY(TREE);
         WRITELN;
         WRITELN('    LOG LIKELIHOOD = ',BESTREE.LIKLI:9:4);
         END;
      WRITELN;
      WRITELN('    TOTAL TREES INSPECTED = ',
                TOTALTREES);
   END; (*MAKETREE*)
BEGIN (*MAIN PROGRAM*)
   RESET (INFILE);
   WRITELN;
   WRITELN(' DNA MAXIMUM LIKELIHOOD ALGORITHM, version 2.1');
   WRITELN;
   MAKETREE (INFILE);
END. (*MAIN PROGRAM*)
