PROGRAM Farris (infile, OUTPUT);
   (* version 2.0 *)
CONST width = 73;  (* form width *)
   nmlngth = 10;   (* length of a species name *)
   maxspp = 50;    (* maximum number of species *)
   maxchars = 190; (* maximum number of characters *)
   maxrange = 190; (* maximum of maxspp, maxchars *)
   bits = 30;      (* number of elements in a pascal set *)
   wrds = 7;       (* bits * wrds >= maxrange *)
TYPE bitrange = 1..bits;
   bitset = SET OF bitrange;
   stateset = ARRAY [1..wrds] OF bitset;
   setarray = ARRAY [1..maxrange] OF stateset;
   name = RECORD
      tekst : ARRAY [1..nmlngth] OF CHAR;
      length : INTEGER;
      END;
   namearray = array [1..maxspp] OF name;

VAR infile : TEXT;
   m, j : stateset;
   chrsets, sppsets : setarray;
   ancestor : ARRAY[1..maxchars] OF CHAR;
   names : namearray;
   spp, chars, min, max, largest, nextlargest, spmin, charmin : INTEGER;
    (* spp = number of species
       chars = number of characters
       min = minimum number of members in a group
       max = maximum number of members in a group *)
   species, sortopt : BOOLEAN;
    (* true if the monothetic groups' members are species,
       their distinguishing sets' members are characters *)

FUNCTION Field (i : INTEGER) : INTEGER;
   (* returns the field width of a positive integer *)
   VAR f : INTEGER;
BEGIN (* Field *)
   IF i < 10
   THEN f := 1
   ELSE IF i < 100
      THEN f := 2
      ELSE f := 3;
   Field := f;
END; (* Field *)

FUNCTION Word (i : INTEGER) : INTEGER;
   (* maps i to its appropriate word in a stateset *)
BEGIN (* Word *)
   Word := (i-1) DIV bits + 1;
END; (* Word *)

FUNCTION Bit (i : INTEGER) : bitrange;
   (* maps i to the appropriate element in the Word *)
BEGIN (* Bit *)
   Bit := (i-1) MOD bits + 1;
END; (* Bit *)

PROCEDURE Readdata;
   (* Read in the numbers of species, of characters, the options,
      and the species names and character state data *)

   PROCEDURE Numsandopts;
      VAR ch : CHAR;
         optnum, i, j : INTEGER;
         ancopt : BOOLEAN;
   BEGIN (* Numsandopts *)
      WRITELN(OUTPUT);
      READ (infile, spp, chars);
      WRITELN(OUTPUT, ' Monothetic Subsets Program, version 2.0');
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, spp:5, ' species,', chars:5, ' characters');
      WRITELN(OUTPUT);
      sortopt := FALSE;
      spmin := 1;
      charmin := 1;
      optnum := 0;
      ancopt := FALSE;
      FOR i := 1 TO chars DO ancestor[i] := '0';
      WHILE NOT EOLN(infile) DO BEGIN
         READ(infile, ch);
         IF ((ch = 'S') OR (ch='s') OR (ch='C') OR (ch='c')
             OR (ch = 'O') OR (ch='o') OR (ch='0')
              OR (ch = 'A') OR (ch = 'a'))
         THEN CASE ch OF
            'A', 'a' : optnum := optnum + 1;
            'S', 's' : optnum := optnum + 1;
            'C', 'c' : optnum := optnum + 1;
            'O', 'o', '0' : sortopt := TRUE;
            END;
         END;
      READLN(infile);
      FOR i := 1 TO optnum DO BEGIN
         READ(infile, ch);
         CASE ch OF
            'A', 'a' : BEGIN
                       ancopt := TRUE;
                       FOR j := 1 TO nmlngth-1 DO
                          READ(infile, ch);
                       FOR j := 1 TO chars DO BEGIN
                          IF EOLN(infile) THEN READLN(infile);
                          READ(infile, ch);
                          ancestor[j] := ch;
                          END;
                       READLN(infile);
                       END;
            'S', 's' : READLN(infile, spmin);
            'C', 'c' : READLN(infile, charmin);
             END;
         END;
      IF ancopt
      THEN BEGIN
         WRITELN(OUTPUT, ' Ancestral states:');
         FOR i := 1 TO nmlngth+3 DO WRITE(OUTPUT, ' ');
         FOR i := 1 TO chars DO BEGIN
            WRITE(OUTPUT, ancestor[i]);
            IF (i MOD 5) = 0 THEN WRITE(OUTPUT, ' ');
            IF (i MOD 50) = 0
            THEN BEGIN
               WRITELN(OUTPUT);
               FOR j := 1 TO nmlngth+4 DO WRITE(' ');
               END;
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         END;
   END; (* Numsandopts *)

   PROCEDURE Getchrsets;
      (* forms the character sets and prints the input data *)
      VAR i, j, k : INTEGER;
         ch : CHAR;
         badchrs : BOOLEAN;

      FUNCTION Skipjunk : CHAR;
         (* returns the next meaningful character read *)
         VAR fnctn : CHAR;
      BEGIN (* Skipjunk *)
         REPEAT
            READ (infile, fnctn);
         UNTIL ((fnctn = '0') OR (fnctn = '1') OR (fnctn = 'b')
                OR (fnctn = 'B') OR (fnctn = 'p') OR (fnctn = 'P')
                OR (fnctn ='?'));
         Skipjunk := fnctn;
      END; (* Skipjunk *)

   BEGIN (* Getchrsets *)
      WRITELN(OUTPUT, ' Species         Characters');
      WRITELN(OUTPUT, ' -------         ----------');
      WRITELN(OUTPUT);
      badchrs := FALSE;
      FOR i := 1 TO spp DO BEGIN
         WRITE(OUTPUT, ' ');
         FOR j := 1 TO nmlngth DO BEGIN
            READ (infile, names[i].tekst[j]);
            WRITE(OUTPUT, names[i].tekst[j]);
            IF names[i].tekst[j] <> ' '
            THEN names[i].length := j;
            END;
         WRITE(OUTPUT, '  ');
         FOR j := 1 TO Word (chars) DO chrsets [i, j] := [];
         FOR j := 1 TO chars DO BEGIN
            IF EOLN (infile)
            THEN READLN (infile);
            ch := Skipjunk;
            WRITE(OUTPUT, ch);
            IF ch <> ancestor[j]
            THEN chrsets [i, Word (j)] :=
                  chrsets [i, Word (j)] + [Bit (j)]
            ELSE
               badchrs := badchrs OR (ch = 'b') OR (ch = 'B')
                                   OR (ch = 'p') OR (ch='P')
                                   OR (ch = '?');
            IF (j MOD 5) = 0
            THEN WRITE(OUTPUT, ' ');
            IF (j MOD 50) = 0
            THEN BEGIN
               WRITELN(OUTPUT);
               FOR k := 1 TO nmlngth+3 DO WRITE(OUTPUT, ' ');
               END;
            END;
         READLN (infile);
         WRITELN(OUTPUT);
         END;
      WRITELN(OUTPUT);
      IF badchrs
      THEN BEGIN
         WRITELN(OUTPUT, ' some states other than 0 or 1 were encountered:');
         WRITELN(OUTPUT, '   these were set to 0');
         WRITELN(OUTPUT);
         END;
   END; (* Getchrsets *)

BEGIN (* Readdata *)
   Numsandopts;
   Getchrsets;
END; (* Readdata *)

PROCEDURE Invert;
   (* constructs the species sets from the character sets *)
   VAR i, j : INTEGER;
BEGIN (* Invert *)
   FOR i := 1 TO chars DO BEGIN
      FOR j := 1 TO Word (spp) DO sppsets [i, j] := [];
      FOR j := 1 TO spp DO
         IF Bit (i) IN chrsets [j, Word (i)]
         THEN sppsets [i, Word (j)] :=
                 sppsets [i, Word (j)] + [Bit (j)];
      END;
END; (* Invert *)

PROCEDURE Printchr (x : stateset; species : BOOLEAN);
   (* prints the indices of the characters in a stateset *)
VAR i, j, printed : INTEGER;
   first : BOOLEAN;
BEGIN (* Printchr *)
   first := TRUE;
   printed := 18;
   j := 0;
   FOR i := 1 TO chars DO
      IF Bit (i) IN x [Word(i)]
      THEN j := j + 1;
   WRITE(OUTPUT, ' ', j:4, ' characters: ');
   FOR i := 1 TO chars DO
      IF Bit (i) IN x [Word (i)]
      THEN BEGIN
         IF NOT first
         THEN BEGIN
            WRITE(OUTPUT, ',');
            printed := printed + 1;
            END;
         IF printed + Field (i) + 2 <= width
         THEN BEGIN
            WRITE(OUTPUT, ' ');
            printed := printed + Field (i) + 1;
            END
         ELSE BEGIN
            WRITELN(OUTPUT);
            WRITE(OUTPUT, ' ');
            WRITE(OUTPUT, '              ');
            printed := Field (i) + 11;
            END;
         WRITE(OUTPUT, i : Field (i));
         IF first THEN first := FALSE;
         END;
    WRITE(OUTPUT, '.');
    WRITELN(OUTPUT);
END; (* Printchr *)

PROCEDURE Printspp (x : stateset; species : BOOLEAN);
   (* prints the names of the species in a stateset *)
   VAR i, j, k, l, printed : INTEGER;
      first : BOOLEAN;
BEGIN (* Printspp *)
   first := true;
   printed := 18;
   j := 0;
   FOR i := 1 TO spp DO
      IF Bit (i) IN x [Word(i)]
      THEN j := j + 1;
   WRITE(OUTPUT, ' ', j: 4, ' species:  ');
   FOR i := 1 TO spp DO
      IF Bit (i) IN x [Word (i)]
      THEN BEGIN
         IF first THEN WRITE(OUTPUT, ' ');
         IF NOT first
         THEN BEGIN
            WRITE(OUTPUT, ',');
            printed := printed + 1;
            END;
         j := names [i].length;
         IF printed + j + 2 <= width
         THEN BEGIN
            WRITE(OUTPUT, ' ');
            printed := printed + j + 1;
            END
         ELSE BEGIN
            WRITELN(OUTPUT);
            WRITE(OUTPUT, ' ');
            WRITE(OUTPUT, '           ');
            printed := j + 11;
            END;
         FOR k := 1 TO j DO
            WRITE(OUTPUT, names [i].tekst [k]);
         IF first THEN first := false;
         END;
   WRITELN(OUTPUT);
END; (* Printspp *)

FUNCTION Count (x : stateset; species : BOOLEAN) : INTEGER;
   (* returns the number of elements in a stateset *)
   VAR i, range, fnctn, b, s : INTEGER;
BEGIN (* Count *)
   IF species
   THEN range := chars
   ELSE range := spp;
   fnctn := 0;
   b := 0;
   s := 1;
   FOR i := 1 TO range DO BEGIN
      b := b + 1;
      IF b in x[s]
      THEN fnctn := fnctn + 1;
      IF b = bits
      THEN BEGIN
         b := 0;
         s := s + 1;
         END;
      END;
   Count := fnctn;
END; (* Count *)

FUNCTION Subset (x, y : stateset; species : BOOLEAN) : BOOLEAN;
   (* true if x is contained in y *)
VAR i, range, b, s : INTEGER;
   fnctn : BOOLEAN;
BEGIN (* Subset *)
   IF species
   THEN range := spp
   ELSE range := chars;
   fnctn := TRUE;
   b := 0;
   s := 1;
   FOR i := 1 TO range DO BEGIN
      b := b + 1;
      IF b IN x[s]
      THEN IF NOT (b IN y[s])
         THEN fnctn := FALSE;
      IF b = bits
      THEN BEGIN
         b := 0;
         s := s + 1;
         END;
      END;
   Subset := fnctn;
END; (* Subset *)

FUNCTION Equal (x, y : stateset; species : BOOLEAN) : BOOLEAN;
   (* true if x and y contain the same members *)
BEGIN (* Equal *)
   Equal := (Subset (x, y, species))
              AND (Subset (y, x, species));
END; (* Equal *)

PROCEDURE Intersect (x : stateset; VAR y : stateset;
                     species : BOOLEAN);
   (* sets y to the intersection of all statesets
      indexed by values in x *)
VAR i, j, range1, range2, b, s : INTEGER;
   sets : setarray;
BEGIN (* Intersect *)
   IF species
   THEN BEGIN
      range1 := spp;
      range2 := chars;
      sets := sppsets
      END
   ELSE BEGIN
      range1 := chars;
      range2 := spp;
      sets := chrsets;
      END;
   FOR i := 1 TO Word (range1) DO y [i] := [];
   IF NOT (Count (x, species) = 0)
   THEN BEGIN
      b := 0;
      s := 1;
      FOR i := 1 TO range1 DO BEGIN
         b := b + 1;
         y [s] := y [s] + [b];
         IF b = bits
         THEN BEGIN
            b := 0;
            s := s + 1;
            END;
         END;
      b := 0;
      s := 1;
      FOR i := 1 TO range2 DO BEGIN
         b := b + 1;
         IF b IN x[s]
         THEN FOR j := 1 TO Word (range1) DO
            y [j] := y [j] * sets [i, j];
         IF b = bits
         THEN BEGIN
            b := 0;
            s := 1;
            END;
         END;
      END;
END; (* Intersect *)

FUNCTION Check (x, y : stateset; species : BOOLEAN) : BOOLEAN;
   (* true if y is contained in the stateset
      of any index in x *)
   VAR i, range, b, s : INTEGER;
      sets : setarray;
      fnctn : BOOLEAN;
BEGIN (* Check *)
   fnctn := FALSE;
   IF species
   THEN BEGIN
      range := chars;
      sets := sppsets;
      END
   ELSE BEGIN
      range := spp;
      sets := chrsets;
      END;
   b := 0;
   s := 1;
   FOR i := 1 TO range DO BEGIN
      b := b + 1;
      IF b IN x[s]
      THEN IF Subset (y, sets [i], species)
         THEN fnctn := true;
      IF b = bits
      THEN BEGIN
         b := 0;
         s := s + 1;
         END;
      END;
   Check := fnctn;
END; (* Check *)

PROCEDURE Gen (m, j : stateset; k, min, max : INTEGER;
               species : BOOLEAN);
   (* recursively produces all monothetic groups *)
   VAR i, range, cnt, n, used, available : INTEGER;
      l, c1, c2, c3 : stateset;
      done : BOOLEAN;
BEGIN (* Gen *)
   IF species
   THEN range := chars
   ELSE range := spp;
   Intersect (j, c2, species);
   used := Count(j, species);
   available := 0;
   FOR n := k TO range DO
      IF NOT(Bit(n) IN j[Word(n)])
      THEN available := available + 1;
   i := k;
   WHILE (i <= range) AND (available + used >= min) DO BEGIN
      IF NOT (Bit(i) IN j[Word(i)])
      THEN BEGIN
         available := available - 1;
         l := j;
         l [Word (i)] := l [Word (i)] + [Bit (i)];
         Intersect (l, c1, species);
         cnt := Count (c1, NOT species);
         IF cnt > 0
         THEN IF NOT Check (m, c1, species)
            THEN BEGIN
               IF NOT Equal (c1, c2, species)
               THEN BEGIN
                  species := NOT species;
                  Intersect (c1, c3, species);
                  species := NOT species;
                  cnt := Count (c3, species);
                  IF cnt > largest
                  THEN largest := cnt;
                  IF (cnt >= min) AND (cnt <= max)
                  THEN IF species
                     THEN BEGIN
                        Printchr (c3, species);
                        Printspp (c1, species);
                        WRITELN(OUTPUT);
                        END
                  ELSE BEGIN
                     Printspp (c3, species);
                     Printchr (c1, species);
                     WRITELN(OUTPUT);
                     END;
                  END;
               IF cnt < max
               THEN Gen (m, c3, i+1, min, max, species);
               m [Word (i)] := m [Word (i)] + [Bit (i)];
               END;
         END;
      i := i + 1;
      END;
END; (* Gen *)

PROCEDURE Chlist;
   VAR i, k : INTEGER;
BEGIN (* Chlist *)
   FOR i := 1 TO wrds DO BEGIN
      m [i] := [];
      j [i] := [];
      END;
   k := 1;
   species := TRUE;
   min := charmin;
   max := chars;
   WRITE(OUTPUT, ' ');
   FOR i := 1 TO width-1 DO WRITE(OUTPUT, '-');
   WRITELN(OUTPUT);
   WRITELN(OUTPUT);
   WRITE(OUTPUT, ' groups');
   IF charmin > 1
   THEN WRITE(OUTPUT, ' with at least', charmin:3, ' characters');
   WRITELN(OUTPUT, ' by order of characters ');
   WRITELN(OUTPUT);
   largest := 0;
   Gen (m, j, k, min, max, species);
   WRITE(OUTPUT, ' ');
   FOR i := 1 TO width-1 DO WRITE(OUTPUT, '-');
   IF sortopt
   THEN BEGIN
      WRITELN(OUTPUT);
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, ' groups by number of characters ');
      WRITELN(OUTPUT);
      i := largest;
      WHILE i > 0 DO BEGIN
         min := i;
         max := i;
         WRITELN(OUTPUT, ' groups containing ',i:3,' characters');
         WRITELN(OUTPUT, ' ------ ---------- --- ----------');
         WRITELN(OUTPUT);
         Gen (m, j, k, min, max, species);
         i := i - 1;
         END;
      WRITE(OUTPUT, ' ');
      FOR i := 1 TO width-1 DO WRITE(OUTPUT, '-');
      END;
END; (* Chlist *)

PROCEDURE Splist;
   VAR i, k : INTEGER;
BEGIN (* Splist *)
   WRITELN(OUTPUT);
   WRITELN(OUTPUT);
   WRITE(OUTPUT, ' groups');
   IF spmin > 1
   THEN WRITE(OUTPUT, ' with at least', spmin:3, ' species');
   WRITELN(OUTPUT, ' by order of species');
   WRITELN(OUTPUT);
   species := FALSE;
   max := spp;
   min := spmin;
   largest := 0;
   k := 1;
   Gen (m, j, k, min, max, species);
   WRITE(OUTPUT, ' ');
   FOR i := 1 TO width DO WRITE(OUTPUT, '-');
   IF sortopt
   THEN BEGIN
      WRITELN(OUTPUT);
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, ' groups by number of species');
      WRITELN(OUTPUT);
      i := largest;
      WHILE i > 0 DO BEGIN
         min := i;
         max := i;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' groups containing ',i:3,' species');
         WRITELN(OUTPUT, ' ------ ---------- --- -------');
         WRITELN(OUTPUT);
         Gen (m, j, k, min, max, species);
         i := i - 1;
         END;
      WRITE(OUTPUT, ' ');
      FOR i := 1 TO width-1 DO WRITE(OUTPUT, '-');
      WRITELN(OUTPUT);
      END;
END; (* Splist *)

BEGIN (* Farris *)
RESET(infile);
Readdata;
Invert;
Chlist;
Splist;
(* may need here - close(infile) *)
END. (* Farris *)
