PROGRAM Newcon(infile, OUTPUT);
   (* version 2.1 *)

CONST maxsp = 20;   (* maximum number of species *)
   maxsp2 = 38;     (* 2 x maxsp - 2 *)
   maxchr = 20;     (* maximum number of characters *)
   smoothings = 4;  (* number of passes through smoothing algorithm *)
   namelength = 10; (* number of characters max. in species name *)
   iterations = 4;  (* number of cycles through covariance iterations *)

TYPE phenotype = ARRAY [1..maxchr] OF REAL;
   naym = ARRAY[1..namelength] OF CHAR;

VAR infile : TEXT;
   numsp, numsp1, numsp2, numchar : INTEGER;
   x, y : ARRAY[1..maxsp] OF phenotype;
   names : ARRAY[1..maxsp] of naym;
   usertree : BOOLEAN;

   PROCEDURE getinput;

      PROCEDURE getnums;
      BEGIN (* getnums *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Iterative Continuous Characters');
         WRITELN(OUTPUT, '    Maximum Likelihood Method version 2.1');
         WRITELN(OUTPUT);
         READ(infile, numsp, numchar);
         WRITELN(OUTPUT, ' ', numsp:4, ' Populations, ', numchar:4,
           ' Phenotypes');
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 2;
      END; (* getnums *)

      PROCEDURE getoptions;
         VAR ch : CHAR;
      BEGIN (* getoptions *)
         usertree := FALSE;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            IF ch <> ' '
            THEN BEGIN
               IF NOT( (ch >= 'A') AND (ch <= 'Z') )
               THEN ch := CHR(ORD(ch)+ORD('Z')-ORD('z'));
               IF ch = 'U'
               THEN usertree := TRUE;
               END;
            END;
      END; (* getoptions *)

      PROCEDURE getdata;
         VAR i, j : INTEGER;
      BEGIN (* getdata *)
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' Name                       Phenotypes');
         WRITELN(OUTPUT, ' ----                       ----------');
         WRITELN(OUTPUT, ' ');
         FOR i := 1 TO numsp DO BEGIN
            WRITE(OUTPUT, ' ');
            READLN(infile);
            FOR j := 1 TO namelength DO BEGIN
               READ(infile, names[i, j]);
               WRITE(OUTPUT, names[i, j]);
               END;
            FOR j := 1 TO numchar DO BEGIN
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, x[i, j]);
               WRITE(OUTPUT, x[i, j]:10:5);
               IF (j MOD 6) = 0 THEN BEGIN
                  WRITELN(OUTPUT);
                  WRITE(OUTPUT, '             ');
                  END;
               END;
            WRITELN(OUTPUT);
            END;
      END; (* getdata *)

   BEGIN (* getinput *)
      getnums;
      getoptions;
      getdata;
   END; (* getinput *)

   PROCEDURE maketree;

      TYPE nodeptr = ^node;
         node = RECORD
            next, back : nodeptr;
            tip        : BOOLEAN;
            number     : INTEGER;
            view       : phenotype;
            nayme      : naym;
            v, deltav  : REAL;
            bigv, dist : REAL;
            xcoord,
             ycoord    : INTEGER;
            END;
         tree = RECORD
            nodep      : ARRAY[1..maxsp2] OF nodeptr;
            likelihood : REAL;
            start      : nodeptr;
            END;
         transmatrix = ARRAY[1..maxchr] of phenotype;

      VAR curtree, priortree, bestree : tree;
         nextsp, numtrees, which, its : INTEGER;
         succeeded : BOOLEAN;
         covariance, transform : transmatrix;

      PROCEDURE setuptree (VAR a : tree);
        VAR i, j : INTEGER;
           p, q : nodeptr;
      BEGIN (* setuptree *)
         WITH a DO BEGIN
            FOR i := 1 TO numsp DO BEGIN
               NEW(nodep[i]);
               nodep[i]^.tip := TRUE;
               nodep[i]^.number := i;
               END;
            FOR i := numsp1 TO numsp2 DO BEGIN
               q := NIL;
               FOR j := 1 TO 3 DO BEGIN
                  NEW(p);
                  p^.tip := FALSE;
                  p^.number := i;
                  p^.next := q;
                  q := p;
                  END;
               p^.next^.next^.next := p;
               nodep[i] := p;
               END;
            likelihood := -99999.0;
            start := nodep[1];
            END;
      END; (* setuptree *)

      PROCEDURE hookup (p, q : nodeptr);
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      FUNCTION evaluate( VAR t : tree ) : REAL;
         VAR sum : REAL;

         PROCEDURE sumlikely (p, q : nodeptr);
            VAR i : INTEGER;
               sumsq, vee : REAL;
         BEGIN (* sumlikely *)
            IF NOT p^.tip THEN sumlikely (p^.next^.back, p^.next^.next^.back);
            IF NOT q^.tip THEN sumlikely (q^.next^.back, q^.next^.next^.back);
            sumsq := 0.0;
            FOR i := 1 TO numchar DO
               sumsq := sumsq + sqr(p^.view[i] - q^.view[i]);
            vee := p^.v + p^.deltav + q^.deltav;
            IF p^.number <> q^.back^.number THEN vee := vee + q^.v;
            sum := sum - numchar*ln(vee)/2 - sumsq/(2*vee);
         END; (* sumlikely *)

      BEGIN (* evaluate *)
         sum := 0.0;
         sumlikely (t.start, t.start^.back);
         t.likelihood := sum;
         evaluate := sum;
      END; (* evaluate *)

      PROCEDURE update (p : nodeptr);
         VAR q : nodeptr;
            negatives : BOOLEAN;

         PROCEDURE makedists (p : nodeptr);
            VAR i : INTEGER;

            FUNCTION distance (p, q : nodeptr) : REAL;
               VAR i : INTEGER;
                  sum : REAL;
            BEGIN (* distance *)
               sum := 0.0;
               FOR i := 1 TO numchar DO
                  sum := sum + sqr(p^.view[i] - q^.view[i]);
               distance := sum;
            END; (* distance *)

         BEGIN (* makedists *)
            FOR i := 1 TO 3 DO BEGIN
               q := p^.next;
               p^.dist := distance (p^.back, q^.back);
               p := q;
               END;
         END; (* makedists *)

         PROCEDURE makebigv (p : nodeptr);
            VAR i : INTEGER;
               temp, q, r : nodeptr;
         BEGIN (* makebigv *)
            q := p^.next;
            r := q^.next;
            negatives := FALSE;
            FOR i := 1 TO 3 DO BEGIN
               p^.bigv := (p^.dist + r^.dist - q^.dist)/(2*numchar);
               IF p^.bigv < p^.back^.deltav THEN negatives := TRUE;
               temp := p;
               p := q;
               q := r;
               r := temp;
               END;
         END; (* makebigv *)

         PROCEDURE correctv (p : nodeptr);
            VAR q, r, temp : nodeptr;
               i, j : INTEGER;
               f1, f2, vtot : REAL;
         BEGIN (* correctv *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO smoothings DO BEGIN
               FOR j := 1 TO 3 DO BEGIN
                  vtot := q^.bigv + r^.bigv;
                  f1 := q^.bigv/vtot;
                  f2 := 1.0 - f1;
                  p^.bigv := (f1*r^.dist + f2*p^.dist - f1*f2*q^.dist)/numchar;
                  p^.bigv := p^.bigv - vtot*f1*f2;
                  IF p^.bigv < p^.back^.deltav
                  THEN p^.bigv := p^.back^.deltav;
                  temp := p;
                  p := q;
                  q := r;
                  r := temp;
                  END;
               END;
         END; (* correctv *)

         PROCEDURE littlev (p : nodeptr);
            VAR i : INTEGER;
         BEGIN (* littlev *)
            FOR i := 1 TO 3 DO BEGIN
               WITH p^ DO BEGIN
                  v := bigv - back^.deltav;
                  back^.v := v;
                  END;
               p := p^.next;
               END;
         END; (* littlev *)

         PROCEDURE nuview (p : nodeptr);
            VAR i, j : INTEGER;
               q, r, a, b, temp : nodeptr;
               v1, v2, vtot, f1, f2 : REAL;
         BEGIN (* nuview *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO 3 DO BEGIN
               a := q^.back;
               b := r^.back;
               v1 := q^.bigv;
               v2 := r^.bigv;
               vtot := v1 + v2;
               f1 := v2/vtot;
               f2 := v1/vtot;
               FOR j := 1 TO numchar DO
                  p^.view[j] := f1*a^.view[j] + f2*b^.view[j];
               p^.deltav := v1*v2/vtot;
               temp := p;
               p := q;
               q := r;
               r := temp;
               END;
         END; (* nuview *)

      BEGIN (* update *)
         IF NOT p^.TIP
         THEN BEGIN
            makedists (p);
            makebigv (p);
            IF negatives THEN correctv (p);
            littlev (p);
            nuview (p);
            END;
      END; (* update *)

      PROCEDURE smooth (p : nodeptr);
      BEGIN (* smooth *)
         IF NOT p^.tip
         THEN BEGIN
            update (p);
            smooth (p^.next^.back);
            smooth (p^.next^.next^.back);
            END;
      END; (* smooth *)

      PROCEDURE insert (p, q: nodeptr);
         VAR i : INTEGER;
      BEGIN (* insert *)
         hookup (p^.next^.next, q^.back);
         hookup (p^.next, q);
         FOR i :=1 TO smoothings DO BEGIN
            smooth (p);
            smooth (p^.back);
            END;
      END; (* insert *)

      PROCEDURE copy ( VAR a, b : tree);
         VAR i : INTEGER;

         PROCEDURE copynode( VAR c, d : nodeptr);
            VAR necks : nodeptr;
         BEGIN (* copynode *)
            necks := d^.next;
            d^ := c^;
            d^.next := necks;
         END; (* copynode *)

      BEGIN (* copy *)
         FOR i := 1 TO numsp DO b.nodep[i]^ := a.nodep[i]^;
         FOR i := numsp1 TO numsp2 DO BEGIN
            copynode (a.nodep[i], b.nodep[i]);
            copynode (a.nodep[i]^.next, b.nodep[i]^.next);
            copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
            END;
         b.likelihood := a.likelihood;
         b.start := a.start;
      END; (* copy *)

      PROCEDURE setuptip (m : INTEGER; VAR t : tree);
      BEGIN (* setuptip *)
         WITH t.nodep[m]^ DO BEGIN
            view := y[m];
            nayme := names[m];
            deltav := 0.0;
            v := 0.0;
            END;
      END; (* setuptip *)

      PROCEDURE buildnewtip (m : INTEGER; VAR t : tree);
         VAR p : nodeptr;
      BEGIN (* buildnewtip *)
         setuptip (m, t);
         p := t.nodep[m + numsp - 2];
         hookup (t.nodep[m], p);
      END; (* buildnewtip *)

      PROCEDURE buildsimpletree (VAR t : tree);
      BEGIN (* buildsimpletree *)
         setuptip (1, t);
         setuptip (2, t);
         hookup (t.nodep[1], t.nodep[2]);
         buildnewtip (3, t);
         insert(t.nodep[3]^.back, t.nodep[1]);
      END; (* buildsimpletree *)

      PROCEDURE addtraverse (p, q : nodeptr);
      BEGIN (* addtraverse *)
         insert(p, q);
         numtrees := numtrees + 1;
         IF evaluate(curtree) > bestree.likelihood
         THEN copy (curtree, bestree);
         copy (priortree, curtree);
         IF NOT q^.tip
         THEN BEGIN
            addtraverse (p, q^.next^.back);
            addtraverse (p, q^.next^.next^.back);
            END;
      END; (* addtraverse *)

      PROCEDURE rearrange (p, q : nodeptr);

         PROCEDURE remove ( VAR p, q : nodeptr);
         BEGIN (* remove *)
            q := p^.next^.back;
            hookup (q, p^.next^.next^.back);
            p^.next^.back := NIL;
            p^.next^.next^.back := NIL;
            update (q);
            update (q^.back);
         END; (* remove *)

         PROCEDURE tryrearr (p, q : nodeptr);
            VAR r : nodeptr;
         BEGIN (* tryrearr *)
            r := q;
            remove (p, r);
            insert (p, q);
            numtrees := numtrees + 1;
            IF evaluate(curtree) > bestree.likelihood
            THEN BEGIN
               succeeded := TRUE;
               copy (curtree, bestree);
               END
            ELSE copy (bestree, curtree);
         END; (* tryrearr *)

      BEGIN (* rearrange *)
         IF NOT succeeded
         THEN BEGIN
            IF (NOT p^.tip) AND (NOT q^.tip)
            THEN BEGIN
               tryrearr (p, q^.next);
               tryrearr (p, q^.next^.next);
               END;
            IF NOT q^.tip
            THEN BEGIN
               rearrange (q, q^.next^.back);
               rearrange (q, q^.next^.next^.back);
               END;
            END;
      END; (* rearrange *)

      PROCEDURE printree;
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + over;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.next^.back);
               p^.xcoord := p^.next^.back^.xcoord;
               coordinates (p^.next^.next^.back);
               p^.ycoord := p^.next^.next^.back^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := curtree.start^.back;
            WRITE(OUTPUT, ' ');
            REPEAT
               IF p^.next^.back^.ycoord >= i
               THEN p := p^.next^.back
               ELSE BEGIN
                  q := p^.next^.next^.back;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF p^.number > 9
                     THEN WRITE(OUTPUT, p^.number:2)
                     ELSE WRITE(OUTPUT, p^.number:1, '-');
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO namelength DO WRITE(OUTPUT, p^.nayme[j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         coordinates (curtree.start^.back);
         WRITELN(OUTPUT, ' ');
         FOR i := 1 TO tipy-down DO drawline (i);
         FOR i := 1 TO down DO WRITELN(OUTPUT, ' !');
         WRITE(OUTPUT, ' ');
         FOR i := 1 TO namelength DO WRITE(OUTPUT, curtree.start^.nayme[i]);
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE summarize;
         VAR chilow, chihigh : REAL;

         PROCEDURE describe (p : nodeptr);
            VAR i : INTEGER;
               q : nodeptr;
               bigv, delta : REAL;
         BEGIN (* describe *)
            q := p^.back;
            WRITE(OUTPUT, ' ');
            IF q^.tip
            THEN FOR i := 1 TO namelength DO WRITE(OUTPUT, q^.nayme[i])
            ELSE WRITE(OUTPUT, q^.number:4, '      ');
            WRITE(OUTPUT, '    ');
            IF p^.tip
            THEN FOR i := 1 TO namelength DO WRITE(OUTPUT, p^.nayme[i])
            ELSE WRITE(OUTPUT, p^.number:4, '      ');
            WRITE(OUTPUT, q^.v:15:5);
            delta := p^.deltav + p^.back^.deltav;
            bigv := p^.v + delta;
            WRITELN(OUTPUT, '   (', chilow*bigv-delta:12:5, ',',
               chihigh*bigv-delta:12:5,')');
            IF NOT p^.TIP
            THEN BEGIN
               describe (p^.next^.back);
               describe (p^.next^.next^.back);
               END;
         END; (* describe *)

      BEGIN (* summarize *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' remember, this is an unrooted tree!');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Ln Likelihood = ', curtree.likelihood:11:5);
         IF NOT usertree
         THEN BEGIN
            WRITELN(OUTPUT, ' ');
            WRITELN(OUTPUT, ' Examined ',numtrees:4, ' trees');
            END;
         IF numchar = 1
         THEN BEGIN
            chilow := 0.000982;
            chihigh := 5.02389;
            END
         ELSE IF numchar = 2
            THEN BEGIN
               chilow := 0.05064;
               chihigh := 7.3777;
               END
            ELSE BEGIN
               chilow := 1.0 - 2.0/(9*numchar);
               chihigh := chilow;
               chilow := chilow - 1.95996 * sqrt(2.0/(9*numchar));
               chihigh := chihigh + 1.95996 * sqrt(2.0/(9*numchar));
               chilow := chilow * chilow * chilow;
               chihigh := chihigh * chihigh * chihigh;
               END;
         WRITELN(OUTPUT, ' ');
         WRITE(OUTPUT, '  Between        And            Length');
         WRITELN(OUTPUT, '      Approx. Confidence Limits');
         WRITE(OUTPUT, '  -------        ---            ------');
         WRITELN(OUTPUT, '      ------- ---------- ------');
         describe (curtree.start^.back);
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' ');
      END; (* summarize *)

      PROCEDURE treeread;
         VAR ch : CHAR;
            p : nodeptr;
            i, nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (p : nodeptr);
            VAR q : nodeptr;
               n : INTEGER;
         BEGIN (* addelement *)
            REPEAT
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := curtree.nodep[nextnode];
               hookup(p, q);
               addelement (q^.next);
               findch (',');
               addelement (q^.next^.next);
               findch (')');
               READ(infile, ch);
               END
            ELSE BEGIN
               n := 0;
               REPEAT
                  n := 10*n + ORD(ch) - ORD('0');
                  READ(infile, ch);
               UNTIL NOT((ORD(ch)>=ORD('0')) AND (ORD(ch)<=ORD('9')));
               hookup (curtree.nodep[n], p);
               IF curtree.start^.number > n
               THEN curtree.start := curtree.nodep[n];
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         curtree.start := curtree.nodep[numsp];
         REPEAT
            READ(infile, ch);
         UNTIL ch <> ' ';
         IF ch = '('
         THEN BEGIN
            nextnode := numsp + 1;
            p := curtree.nodep[nextnode];
            FOR i := 1 TO 2 DO BEGIN
               addelement (p);
               p := p^.next;
               findch (',');
               END;
            addelement (p);
            FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
            END;
         READLN(infile);
      END; (* treeread *)

      PROCEDURE treevaluate;
         VAR p : nodeptr;
            dummy : REAL;

         PROCEDURE nodeinit (p : nodeptr);
            VAR q, r : nodeptr;
               i : INTEGER;
         BEGIN (* nodeinit *)
            IF NOT p^.tip
            THEN BEGIN
               q := p^.next^.back;
               r := p^.next^.next^.back;
               nodeinit (q);
               nodeinit (r);
               FOR i := 1 TO numchar DO
                  p^.view[i] := 0.5 * q^.view[i] + 0.5 * r^.view[i];
               END;
         END; (* nodeinit *)

      BEGIN (* treevaluate *)
         nodeinit (curtree.start);
         dummy := evaluate (curtree);
      END; (* treevaluate *)

      PROCEDURE inittransform;
         VAR i, j : INTEGER;
      BEGIN (* inittransform *)
         FOR i := 1 TO numchar DO
            FOR j := 1 TO numchar DO
               IF i = j
               THEN transform[i,i] := 1.0
               ELSE transform[i,j] := 0.0;
      END; (* inittransform *)

      PROCEDURE transfrmdata;
         VAR sum : REAL;
            i, j, k : INTEGER;
      BEGIN (* transfrmdata *)
         FOR i := 1 TO numsp DO
            FOR j := 1 TO numchar DO BEGIN
               sum := 0.0;
               FOR k := 1 TO numchar DO
                  sum := sum + x[i,k] * transform[j,k];
               y[i,j] := sum;
               END;
      END; (* transfrmdata *)

      PROCEDURE covarmatrix (VAR t : tree);
         VAR i, j : INTEGER;

         PROCEDURE sumcovar (p, q : nodeptr);
            VAR i, j : INTEGER;
               vee : REAL;
         BEGIN (* sumcovar *)
            IF NOT p^.tip THEN sumcovar (p^.next^.back,
                                             p^.next^.next^.back);
            IF NOT q^.tip THEN sumcovar (q^.next^.back,
                                             q^.next^.next^.back);
            vee := p^.v + p^.deltav + q^.deltav;
            IF p^.back^.number <> q^.number THEN vee := vee + q^.v;
            FOR i := 1 TO numchar DO
               FOR j := 1 TO i DO
                  covariance[i,j] := covariance[i,j] + (p^.view[i]-q^.view[i])
                     * (p^.view[j]-q^.view[j])/vee;
         END; (* sumcovar *)

      BEGIN (* covarmatrix *)
         FOR i := 1 TO numchar DO
            FOR j := 1 TO i DO
               covariance[i,j] := 0.0;
         sumcovar (curtree.start, curtree.start^.back);
         FOR i := 1 TO numchar-1 DO
            FOR j := i+1 TO numchar DO
               covariance[i,j] := covariance[j,i];
         FOR i := 1 TO numchar DO
            FOR j := 1 TO numchar DO
               covariance[i,j] := covariance[i,j]/numchar;
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' Covariance Matrix');
         WRITELN(OUTPUT, ' ');
         FOR i := 1 TO numchar DO BEGIN
            FOR j := 1 TO numchar DO BEGIN
               WRITE(OUTPUT, covariance[i,j]:10:5);
               IF (j MOD 7) = 0
               THEN BEGIN
                  WRITELN(OUTPUT);
                  WRITE(OUTPUT, '   ');
                  END;
                END;
            IF (numchar MOD 7) <> 0
            THEN WRITELN(OUTPUT);
            END;
      END; (* covarmatrix *)

      PROCEDURE maketransform;
         VAR temp : transmatrix;

         PROCEDURE squaroot;
            VAR sum : REAL;
               i, j, k : INTEGER;
         BEGIN (* squaroot *)
            FOR k := 1 TO numchar DO BEGIN
               FOR j := 1 TO k DO BEGIN
                  sum := 0.0;
                  FOR i := 1 TO j-1 DO sum := sum + covariance[k,i]
                                                    * covariance[j,i];
                  temp[k,j] := covariance[k,j] - sum;
                  IF k = j
                  THEN IF temp[k,j] <= 0.0
                      THEN temp[k,j] := 0.0
                      ELSE temp[k,j] := sqrt(temp[k,j])
                  ELSE IF temp[j,j] <= 0.0
                     THEN temp[k,j] := 0.0
                     ELSE temp[k,j] := temp[k,j]/temp[j,j];
                  END;
               END;
         END; (* squaroot *)

         PROCEDURE invertit;
            VAR i, j, k : INTEGER;
               sum : REAL;
         BEGIN (* invertit *)
            FOR j := 1 TO numchar DO
               FOR i := j TO numchar DO BEGIN
                  sum := 0.0;
                  IF i = j THEN sum := 1.0;
                  FOR k := j TO i-1 DO
                     sum := sum - temp[i,k] * covariance[k,j];
                  IF ABS(temp[i,i]) > 0.000001
                  THEN covariance[i,j] := sum / temp[i,i]
                  ELSE covariance[i,j] := 0.0;
                  END;
         END; (* invertit *)

         PROCEDURE multiply;
            VAR sum : REAL;
               i, j, k : INTEGER;
         BEGIN (* multiply *)
            FOR i := 1 TO numchar DO
               FOR k := 1 TO numchar DO BEGIN
                  sum := 0.0;
                  FOR j := 1 TO i DO
                     sum := sum + covariance[i,j] * transform[j,k];
                  temp[i,k] := sum;
                  END;
            FOR i := 1 TO numchar DO
               FOR j := 1 TO numchar DO
                  transform[i,j] := temp[i,j];
         END; (* multiply *)

         PROCEDURE printit;
            VAR i, j : INTEGER;
         BEGIN (* printit *)
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, ' Transform Becomes:');
            WRITELN(OUTPUT);
            FOR i := 1 TO numchar DO BEGIN
               FOR j := 1 TO i DO BEGIN
                  IF (j MOD 7) = 0
                  THEN BEGIN
                     WRITELN(OUTPUT);
                     WRITE(OUTPUT, '  ');
                     END;
                  WRITE(OUTPUT, transform[i,j]:10:5);
                  END;
               WRITELN(OUTPUT);
               END;
          END; (* printit *)

      BEGIN (* maketransform *)
         squaroot;
         invertit;
         multiply;
         printit;
      END; (* maketransform *)

   BEGIN (* maketree *)
      IF usertree
      THEN BEGIN
         setuptree (curtree);
         FOR which := 1 TO numsp DO setuptip (which, curtree);
         READLN(infile, numtrees);
         FOR which := 1 TO numtrees DO BEGIN
            treeread;
            treevaluate;
            printree;
            summarize;
            END;
         END
      ELSE BEGIN
         setuptree (curtree);
         setuptree (priortree);
         setuptree (bestree);
         inittransform;
         FOR its := 1 TO iterations DO BEGIN
            bestree.likelihood := -99999.0;
            transfrmdata;
            buildsimpletree (curtree);
            numtrees := 1;
            nextsp := 4;
            WHILE nextsp <= numsp DO BEGIN
               buildnewtip (nextsp, curtree);
               copy (curtree, priortree);
               bestree.likelihood := -99999.0;
               addtraverse (curtree.nodep[nextsp]^.back, curtree.start^.back);
               copy (bestree, curtree);
               succeeded := TRUE;
               IF nextsp > 3
               THEN WHILE succeeded DO BEGIN
                  succeeded := FALSE;
                  rearrange (curtree.start, curtree.start^.back);
                  END;
               printree;
               summarize;
               nextsp := nextsp + 1;
               END;
            covarmatrix (curtree);
            maketransform;
            END;
         END;
   END; (* maketree *)

BEGIN (* main program *)
   RESET(infile);
   getinput;
   maketree;
END.
